#import "BWStitchNoteDocument.h"
#import "BWDocument-Undo.h"
#import "BWStitchGroup.h"
#import "BWFabricView.h"

@implementation BWStitchNoteDocument(UndoStuff)


// stash away model settings that are controlled by continuous controls.
// This way we can let the controls change the model values, and still
// be able to undo the  last value as the user drags the control around

- (void) rememberUndoInformation
{
    // snarf original values for undo using the coalescing of notifications
    originalFabricWidth = [group fabricWidth];
    originalFabricSpacing = [group fabricSpacing];
    originalStitchSize = [group stitchSize];

    originalFabricColor = [[group fabricColor] copy];
    originalBackgroundColor = [[group backgroundColor] copy];
    originalStitchColor = [[group stitchColor] copy];
    originalTextColor = [[group textColor] copy];
    
} // rememberUndoInformation



// set up any state we need to support undo for this document

- (void) setupUndo
{
    [self rememberUndoInformation];

	
    // set up the notifiction queue to coalesce the
    // continuous controls to make a single undo
    // action.
    NSNotificationCenter *center;
    center = [NSNotificationCenter defaultCenter];

    undoQueue = [[NSNotificationQueue alloc]
		    initWithNotificationCenter: center];

    // the notification names here are the ones that are posted by
    // events put into the undoQueue.  Continuous controls generate a
    // bunch of notifications to that undoQueue.  Once the user stops
    // moving the control around, the undoQueue will post its
    // notification to the default notification center, and then the
    // methods referenced below will be invoked to create the actual
    // undo action.

    [center addObserver: self
	    selector: @selector(fabricWidthChanged:)
	    name: @"FabricWidth"
	    object: nil];

    [center addObserver: self
	    selector: @selector(fabricSpacingChanged:)
	    name: @"FabricSpacing"
	    object: nil];

    [center addObserver: self
	    selector: @selector(stitchSizeChanged:)
	    name: @"StitchSize"
	    object: nil];

    [center addObserver: self
	    selector: @selector(fabricColorChanged:)
	    name: @"FabricColor"
	    object: nil];

    [center addObserver: self
	    selector: @selector(backgroundColorChanged:)
	    name: @"BackgroundColor"
	    object: nil];

    [center addObserver: self
	    selector: @selector(stitchColorChanged:)
	    name: @"StitchColor"
	    object: nil];

    [center addObserver: self
	    selector: @selector(textColorChanged:)
	    name: @"TextColor"
	    object: nil];

} // setupUndo



// clean up any messes we've made

- (void) deallocUndo
{
    [undoQueue release];
    [[NSNotificationCenter defaultCenter] removeObserver: self];

} // deallocUndo



// add a new stitch to the group

- (void) addStitchLocation: (BWStitchLocation) location
{
    NSUndoManager *undoManager;
    undoManager = [self undoManager];

    // since we're adding a new stitch to the end of the group, 
    // to undo it involves remove the last stitch
    [[undoManager prepareWithInvocationTarget: self]
	removeLastStitchLocation];

    // actually add it
    [group addStitchLocation: location];

    [undoManager setActionName: @"Add Stitch"];

    [fabricView setNeedsDisplay: YES];

} // addStitchLocation



// lop off the last stitch

- (void) removeLastStitchLocation
{
    BWStitchLocation location;
    location = [group stitchLocationAtIndex: [group count] - 1];

    NSUndoManager *undoManager;
    undoManager = [self undoManager];

    // since we're removing the last stitch, 
    // to undo it involves adding the stitch to the end of the group.
    [[undoManager prepareWithInvocationTarget: self]
	addStitchLocation: location];

    // actually remove it
    [group removeLastStitchLocation];

    [undoManager setActionName: @"Remove Stitch"];

    [fabricView setNeedsDisplay: YES];

    [self updateUI];

} // removeLastStitchLocation



// change whether certain stitch labels are drawn

- (void) setNumberDisplay: (BWStitchGroup_NumberDisplay) value
{
    NSUndoManager *undoManager;
    undoManager = [self undoManager];

    // since we're changing an enum value,
    // to undo it, set it back to its (now) current value
    [[undoManager prepareWithInvocationTarget: self]
	setNumberDisplay: [group numberDisplay]];

    // actually change the display value
    [group setNumberDisplay: value];

    [undoManager setActionName: @"Number Display"];

    [fabricView setNeedsDisplay: YES];

    // update the pop-up if we get invoked by an undo action
    [self updateUI];

} // setNumberDisplay



// change how the text is displayed

- (void) setFontFamily: (NSString *) familyName  size: (int) size
		  bold: (BOOL) bold  italic: (BOOL) italic
{
    NSFontManager *fontManager;
    fontManager = [NSFontManager sharedFontManager];

    // hang on to the old font values so we can use it to undo to
    NSFont *oldFont;
    oldFont = [group textFont];

    NSString *oldFamily;
    oldFamily = [oldFont familyName];
    
    int oldSize;
    oldSize = (int)[oldFont pointSize];
    
    NSFontTraitMask traits;
    traits = [fontManager traitsOfFont: oldFont];;

    BOOL oldBold;
    oldBold = (traits & NSBoldFontMask);

    BOOL oldItalic;
    oldItalic = (traits & NSItalicFontMask);

    // since we're changing a font value
    // to undo it, restore the previous font
    NSUndoManager *undoManager;
    undoManager = [self undoManager];
    
    [[undoManager prepareWithInvocationTarget: self]
	setFontFamily: oldFamily  size: oldSize
	    bold: oldBold  italic: oldItalic];


    // make the new font

    traits = 0;

    if (bold) {
	traits |= NSBoldFontMask;
    }
    if (italic) {
	traits |= NSItalicFontMask;
    }

    NSFont *font;
    font = [fontManager fontWithFamily: familyName
			traits: traits
			weight: 5
			size: size];

    // actually change the font
    [group setTextFont: font];

    [undoManager setActionName: @"Font Change"];

    // update the pop-ups
    [self updateUI];

    [fabricView setNeedsDisplay: YES];

} // setFontFamily


// move all the stitches around

- (void) shiftStitchesBy: (int) number
	       direction: (BWStitchGroup_ShiftDirection) direction
{
    // figure out the opposite direction

    BWStitchGroup_ShiftDirection oppositeDirection;

    switch (direction) {
    case BWStitchGroup_ShiftUp:
	oppositeDirection = BWStitchGroup_ShiftDown;
	break;
    case BWStitchGroup_ShiftDown:
	oppositeDirection = BWStitchGroup_ShiftUp;
	break;
    case BWStitchGroup_ShiftLeft:
	oppositeDirection = BWStitchGroup_ShiftRight;
	break;
    case BWStitchGroup_ShiftRight:
	oppositeDirection = BWStitchGroup_ShiftLeft;
	break;
    default:
	NSLog (@"bad shift direction");
    }
    
    NSUndoManager *undoManager;
    undoManager = [self undoManager];

    // since we're moving everything over by a specified amount,
    // to undo it, just move it the other direction
    [[undoManager prepareWithInvocationTarget: self]
	shiftStitchesBy: number
	direction: oppositeDirection];

    // actually move the stitches around
    [group shiftStitchesBy: number
	   direction: (BWStitchGroup_ShiftDirection) direction];

    [undoManager setActionName: @"Shift Stitches"];

    [fabricView setNeedsDisplay: YES];

} // shiftStitches



// change the area that will get printed

- (void) setPrintableArea: (BWIntRect) printableArea
{
    // get the old printable area so we can undo to it
    BWIntRect oldPrintableArea;
    oldPrintableArea = [group printableArea];

    NSUndoManager *undoManager;
    undoManager = [self undoManager];

    // since we're changing the rectangle of the printable area,
    // to undo it, set it back to its (now) current value
    [[undoManager prepareWithInvocationTarget: self]
	setPrintableArea: oldPrintableArea];

    // actually change the display value
    [group setPrintableArea: printableArea];

    // if the printable area is empty, this must be a 
    // clear operation
    if (oldPrintableArea.size.width <= 0
	|| printableArea.size.height <= 0) {
	[undoManager setActionName: @"Clear Printable Area"];
    } else {
	[undoManager setActionName: @"Set Printable Area"];
    }

    [fabricView setNeedsDisplay: YES];

} // setPrintableArea



// convenience for setPrintableArea: emptyRect

- (void) clearPrintableArea
{
    // get the old printable area so we can undo to it
    BWIntRect printableArea;
    printableArea = [group printableArea];

    NSUndoManager *undoManager;
    undoManager = [self undoManager];

    // since we're changing the rectangle of the printable area,
    // to undo it, set it back to its (now) current value
    [[undoManager prepareWithInvocationTarget: self]
	setPrintableArea: printableArea];

    // actually change the display value
    [group clearPrintableArea];

    [undoManager setActionName: @"Clear Printable Area"];

    [fabricView setNeedsDisplay: YES];

} // clearPrintableArea



// --------------------------------------------------



// Below here are the functions for doing undo for continuous
// controls.  There are three functions for each undoable operation:

// * enqueueXYZChange:
//      the handler method for a slider or color well sets the value or
//      color in the Group object (so that the view will draw the
//      interim value, and then calls this.  A notification is added to
//      our private undoQueue, which will coalesce them into a single
//      notification to the general notification manager during 'idle'
//      time, which is when the user has finished tracking the control

// * XYZChanged:
//      this is the method that gets invoked when the undoQueue above
//      posts its notification during idle time.  This pulls the value
//      out of the UI control, and then invokes

// * setXYZ:
//      does the undo work.  The prepareWithInvocationTarget uses
//      the saved 'original' value as the value to undo to, and then
//      it does the operation on the Group to change the value.
//      Finally it saves off the new value as the 'original' value
//      for use when the user twiddles this control again.


// --------------------------------------------------



// set the fabric width (the weight of the threads that form the grid

- (void) setFabricWidth: (float) width
{
    // don't do any work if we don't have to
    if (width != originalFabricWidth) {

	NSUndoManager *undoManager;
	undoManager = [self undoManager];
	
	// since we're changing the fabric width,
	// undoing it involves restoring it to the old value.
	// This is one of those continuous controls things.
	// the originalFabricWidth was snarfed prior to the user
	// manipulating the control
	[[undoManager prepareWithInvocationTarget: self]
	    setFabricWidth: originalFabricWidth];
	
	// actually change the width
	[group setFabricWidth: width];
	
	// this is the new 'original' value
	originalFabricWidth = width;
	
	[undoManager setActionName: @"Change Fabric Width"];
	
	[self updateUI];
    }
	
} // setFabricWidth



// this is invoked by our notification queue once the user finishes
// dragging the slider around

- (void) fabricWidthChanged: (NSNotification *) notification
{
    float newWidth;
    newWidth = [[notification object] floatValue];

    [self setFabricWidth: newWidth];

} // fabricWidthChanged



// respond to the fabric width slider being manipulated

- (void) enqueueFabricWidthChange: (id) sender
{
    // the caller has already modified the group.  Here we just
    // arrange to be notified later that the width has changed so
    // we can really do the undo operation at that time.

    NSNotification *notification;
    notification = [NSNotification notificationWithName: @"FabricWidth"
				   object: sender];

    [undoQueue enqueueNotification: notification
	       postingStyle: NSPostWhenIdle
	       coalesceMask: (NSNotificationCoalescingOnName
			      | NSNotificationCoalescingOnSender)
	       forModes: nil];

} // enqueueFabricWidthChange



// --------------------------------------------------



// set the fabric spacing, how far apart the threads that form the
// grid are spaced

- (void) setFabricSpacing: (float) spacing
{
    // don't do work if we don't need to
    if (spacing != originalFabricSpacing) {

	NSUndoManager *undoManager;
	undoManager = [self undoManager];
	
	// since we're changing the fabric spacing,
	// undoing it involves restoring it to the old value.
	// This is one of those continuous controls things.
	// the originalFabricSpacing was snarfed prior to the user
	// manipulating the control
	[[undoManager prepareWithInvocationTarget: self]
	    setFabricSpacing: originalFabricSpacing];
	
	// actually change the width
	[group setFabricSpacing: spacing];
	
	// this is the new 'original' value
	originalFabricSpacing = spacing;
	
	[undoManager setActionName: @"Change Fabric Spacing"];
	
	[self updateUI];
    }

} // setFabricSpacing



// this is invoked by our notification queue once the user finishes
// dragging the slider around

- (void) fabricSpacingChanged: (NSNotification *) notification
{
    float newSpacing;
    newSpacing = [[notification object] floatValue];

    [self setFabricSpacing: newSpacing];

} // fabricSpacingChanged



// respond to the fabric width space being manipulated

- (void) enqueueFabricSpacingChange: (id) sender
{
    // the caller has already modified the group.  Here we just
    // arrange to be notified later that the width has changed so
    // we can really do the undo operation at that time.

    NSNotification *notification;
    notification = [NSNotification notificationWithName: @"FabricSpacing"
				   object: sender];

    [undoQueue enqueueNotification: notification
	       postingStyle: NSPostWhenIdle
	       coalesceMask: (NSNotificationCoalescingOnName
			      | NSNotificationCoalescingOnSender)
	       forModes: nil];

} // enqueueFabricSpacingChange



// --------------------------------------------------



// set the stitch size (how thick the drawn stitches are)

- (void) setStitchSize: (float) size
{
    // don't do work if we don't need to
    if (size != originalStitchSize) {

	NSUndoManager *undoManager;
	undoManager = [self undoManager];
	
	// since we're changing the stitch size
	// undoing it involves restoring it to the old value.
	// This is one of those continuous controls things.
	// the originalStitchSize was snarfed prior to the user
	// manipulating the control
	[[undoManager prepareWithInvocationTarget: self]
	    setStitchSize: originalStitchSize];
	
	// actually change the size
	[group setStitchSize: size];
	
	// this is the new 'original' value
	originalStitchSize = size;
	
	[undoManager setActionName: @"Change Stitch Size"];
	
	[self updateUI];
    }

} // setStitchSize



// this is invoked by our notification queue once the user finishes
// dragging the slider around

- (void) stitchSizeChanged: (NSNotification *) notification
{
    float newWidth;
    newWidth = [[notification object] floatValue];

    [self setStitchSize: newWidth];

} // stitchSizeChanged



// respond to the stitch size slider being manipulated

- (void) enqueueStitchSizeChange: (id) sender
{
    // the caller has already modified the group.  Here we just
    // arrange to be notified later that the width has changed so
    // we can really do the undo operation at that time.

    NSNotification *notification;
    notification = [NSNotification notificationWithName: @"StitchSize"
				   object: sender];

    [undoQueue enqueueNotification: notification
	       postingStyle: NSPostWhenIdle
	       coalesceMask: (NSNotificationCoalescingOnName
			      | NSNotificationCoalescingOnSender)
	       forModes: nil];

} // enqueueStitchSizeChange



// --------------------------------------------------



// set the fabric color (the grid that gets drawn)

- (void) setFabricColor: (NSColor *) color
{
    // the color chooser can send us a spurious "color changed"
    // notification  when first opened.  Only make the
    // undo action if we really need to.

    if (![color isEqualTo: originalFabricColor]) {
	
	NSUndoManager *undoManager;
	undoManager = [self undoManager];
	
	// since we're changing the fabric color
	// undoing it involves restoring it to the old value.
	// This is one of those continuous controls things.
	// the originalFabricColor was snarfed prior to the user
	// manipulating the control
	[[undoManager prepareWithInvocationTarget: self]
	    setFabricColor: originalFabricColor];

	// actually change the color
	[group setFabricColor: color];
	
	// this is the new 'original' color
	[originalFabricColor release];
	originalFabricColor = [color copy];

	[undoManager setActionName: @"Change Fabric Color"];

	[self updateUI];
    }

} // setFabricColor



// this is invoked by our notification queue once the user finishes
// dragging the color well around

- (void) fabricColorChanged: (NSNotification *) notification
{
    NSColor *color;
    color = [[notification object] color];

    [self setFabricColor: color];

} // fabricColorChanged



// respond to the fabric color colorWell being manipulated

- (void) enqueueFabricColorChange: (id) sender
{
    // the caller has already modified the group.  Here we just
    // arrange to be notified later that the width has changed so
    // we can really do the undo operation at that time.

    NSNotification *notification;
    notification = [NSNotification notificationWithName: @"FabricColor"
				   object: sender];

    [undoQueue enqueueNotification: notification
	       postingStyle: NSPostWhenIdle
	       coalesceMask: (NSNotificationCoalescingOnName
			      | NSNotificationCoalescingOnSender)
	       forModes: nil];

} // enqueueFabricColorChange



// --------------------------------------------------



// set the background color (the area not drawn by anything else)

- (void) setBackgroundColor: (NSColor *) color
{
    // the color chooser can send us a spurious "color changed"
    // notification  when first opened.  Only make the
    // undo action if we really need to.
    if (![color isEqualTo: originalBackgroundColor]) {
	
	NSUndoManager *undoManager;
	undoManager = [self undoManager];
	
	// since we're changing the background color
	// undoing it involves restoring it to the old value.
	// This is one of those continuous controls things.
	// the originalBackgroundColor was snarfed prior to the user
	// manipulating the control
	[[undoManager prepareWithInvocationTarget: self]
	    setBackgroundColor: originalBackgroundColor];

	// actually change the color
	[group setBackgroundColor: color];

	// this is the new 'original' color
	[originalBackgroundColor release];
	originalBackgroundColor = [color copy];

	[undoManager setActionName: @"Change Background Color"];

	[self updateUI];
    }

} // setBackgroundColor



// this is invoked by our notification queue once the user finishes
// dragging the color well around

- (void) backgroundColorChanged: (NSNotification *) notification
{
    NSColor *color;
    color = [[notification object] color];

    [self setBackgroundColor: color];

} // backgroundColorChanged



// respond to the background color colorwell  been manipulated

- (void) enqueueBackgroundColorChange: (id) sender
{
    // the caller has already modified the group.  Here we just
    // arrange to be notified later that the width has changed so
    // we can really do the undo operation at that time.

    NSNotification *notification;
    notification = [NSNotification notificationWithName: @"BackgroundColor"
				   object: sender];

    [undoQueue enqueueNotification: notification
	       postingStyle: NSPostWhenIdle
	       coalesceMask: (NSNotificationCoalescingOnName
			      | NSNotificationCoalescingOnSender)
	       forModes: nil];

} // enqueueBackgroundColorChange



// --------------------------------------------------



// set the stitch color

- (void) setStitchColor: (NSColor *) color
{
    // the color chooser can send us a spurious "color changed"
    // notification  when first opened.  Only make the
    // undo action if we really need to.
    if (![color isEqualTo: originalStitchColor]) {
	
	NSUndoManager *undoManager;
	undoManager = [self undoManager];
	
	// since we're changing the stitch color
	// undoing it involves restoring it to the old value.
	// This is one of those continuous controls things.
	// the originalStitchColor was snarfed prior to the user
	// manipulating the control
	[[undoManager prepareWithInvocationTarget: self]
	    setStitchColor: originalStitchColor];

	// actually change the color
	[group setStitchColor: color];

	// this is the new 'original' color
	[originalStitchColor release];
	originalStitchColor = [color copy];

	[undoManager setActionName: @"Change Stitch Color"];
	
	[self updateUI];
    }

} // setStitchColor



// this is invoked by our notification queue once the user finishes
// dragging the color well around

- (void) stitchColorChanged: (NSNotification *) notification
{
    NSColor *color;
    color = [[notification object] color];

    [self setStitchColor: color];

} // stitchColorChanged



// respond to the fabric color colorWell being manipulated

- (void) enqueueStitchColorChange: (id) sender
{
    // the caller has already modified the group.  Here we just
    // arrange to be notified later that the width has changed so
    // we can really do the undo operation at that time.

    NSNotification *notification;
    notification = [NSNotification notificationWithName: @"StitchColor"
				   object: sender];

    [undoQueue enqueueNotification: notification
	       postingStyle: NSPostWhenIdle
	       coalesceMask: (NSNotificationCoalescingOnName
			      | NSNotificationCoalescingOnSender)
	       forModes: nil];

} // enqueueStitchColorChange



// --------------------------------------------------



// set the text color

- (void) setTextColor: (NSColor *) color
{
    // the color chooser can send us a spurious "color changed"
    // notification  when first opened.  Only make the
    // undo action if we really need to.
    if (![color isEqualTo: originalTextColor]) {
	
	NSUndoManager *undoManager;
	undoManager = [self undoManager];
	
	// since we're changing the text color
	// undoing it involves restoring it to the old value.
	// This is one of those continuous controls things.
	// the originalTextColor was snarfed prior to the user
	// manipulating the control
	[[undoManager prepareWithInvocationTarget: self]
	    setTextColor: originalTextColor];

	// actually change the color
	[group setTextColor: color];

	// this is the new 'original' color
	[originalTextColor release];
	originalTextColor = [color copy];

	[undoManager setActionName: @"Change Text Color"];
	
	[self updateUI];
    }

} // setTextColor



// this is invoked by our notification queue once the user finishes
// dragging the colorWell around

- (void) textColorChanged: (NSNotification *) notification
{
    NSColor *color;
    color = [[notification object] color];

    [self setTextColor: color];

} // textColorChanged



// respond to the text color colorWell being manipulated

- (void) enqueueTextColorChange: (id) sender
{
    // the caller has already modified the group.  Here we just
    // arrange to be notified later that the width has changed so
    // we can really do the undo operation at that time.

    NSNotification *notification;
    notification = [NSNotification notificationWithName: @"TextColor"
				   object: sender];

    [undoQueue enqueueNotification: notification
	       postingStyle: NSPostWhenIdle
	       coalesceMask: (NSNotificationCoalescingOnName
			      | NSNotificationCoalescingOnSender)
	       forModes: nil];

} // enqueueTextColorChange


// --------------------------------------------------



@end // BWStitchNoteDocument(UndoStuff)


