// BWFabricView.h -- show the "fabric" bacground, and let the user draw
//                   stitches

#import "BWFabricView.h"

#import "BWStitchGroup.h"
#import "BWCursor.h"


// notifications when the user adds a new stitch location,
// or they finish selecting the printable area
NSString *BWFabricViewNewStitchNotification 
    = @"BWFabricView NewStitch";
NSString *BWFabricViewPrintableAreaNotification
    = @"BWFabricView PrintableAreaNotifcation";



@implementation BWFabricView


- (BWIntRect) tweakPrintableArea: (BWIntRect) printableArea
{
    // expand the group's printable area by 1 so that numbers
    // drawn at the end don't get clipped.
    // this is just an approximation since we don't look at the
    // font size and placement here
    
    printableArea.origin.x--;
    printableArea.origin.y--;
    printableArea.size.width += 2;
    printableArea.size.height += 2;

    return (printableArea);
    
} // tweakPrintableArea



// we're being created for printing, rather than for interactive
// abuse

- (id) initForPrintingWithGroup: (BWStitchGroup *) theGroup
{
    // can't init with frame since we don't have a frame yet
    if (self = [super init]) {

	// we need to know who our Model object is
	[self setStitchGroup: theGroup];

	// figure out the area to print
	BWIntRect printableArea;
	printableArea = [group printableArea];
    
	// empty rect, so there isn't an explicit printable area.
	// try to figure out the right thing to print
	if (printableArea.size.width == 0 || printableArea.size.height == 0) {
	    printableArea = [group contentBounds];
	    printableArea = [self tweakPrintableArea: printableArea];
	}
    
	// Determine the location of the top-left cell.
	// We just want the origin of the rectangle, which
	// will become the origin of the printing area
	BWStitchLocation topLeft;
	topLeft = BWMakeStitchLocation (printableArea.origin.x,
					printableArea.origin.y);
    
	NSRect cellRect;
	cellRect = [self rectForLocation: topLeft];
	
	// determine the actual width and height of the area to print
	NSSize printableSize;
	printableSize = [self printableSize];

	// make the rectangle with our origin and size
	NSRect bounds;
	bounds.origin = cellRect.origin;
	bounds.size = printableSize;
    
	// and then account for fabric weave size
	// The +1 fudge factor is for the black rule on either size 
	// of the stitches
	float fabricWidth;
	fabricWidth = [group fabricWidth];
	bounds.origin.x -= (fabricWidth / 2.0) + 1;
	bounds.origin.y -= (fabricWidth / 2.0) + 1;

	// doing both seem necessary.  Not setting the frame size
	// causes a "no pages from the document were selected to be
	// printed" complaint
 	[self setFrameSize: printableSize];
	[self setBounds: bounds];
    }

    return (self);

} // initForPrintingWithGroup



// we've been loaded with connections set up

- (void) awakeFromNib
{
    NSWindow *parentWindow;

    // make sure the window will give us mouse motion events
    parentWindow = [self window];
    [parentWindow setAcceptsMouseMovedEvents: YES];

    // no tracking display yet
    trackX = BWFabricView_NO_TRACK;
    trackY = BWFabricView_NO_TRACK;
    
    // this will draw in a transparent black to darken the cell
    // will break on black backgrounds
    cellHighlightColor = [NSColor colorWithCalibratedRed: 0.0
				  green: 0.0
				  blue: 0.0
				  alpha: 0.5];
    [cellHighlightColor retain];

} // awakeFromNib



// clean up our mess

- (void) dealloc
{
    [textAttributes release];

    // disconnect any observers
    NSNotificationCenter *center;
    center = [NSNotificationCenter defaultCenter];

    [center removeObserver: self
	    name: BWStitchGroup_VisualAttributeChangeNotification
	    object: group];

    [cellHighlightColor release];
    
    [super dealloc];

} // dealloc



// Given a cell location, determine the rectangle in the view

- (NSRect) rectForLocation: (BWStitchLocation) location
{
    float rx, ry;

    // the fabricWidth is for skipping over the solid colored grid
    // lines, and the spacing is the cell's internal size
    float fabricWidth, spacing;
    fabricWidth = [group fabricWidth];
    spacing = [group fabricSpacing];

    // the fabricWidth / 2.0 is there so that the upper-left of the
    // view will be half-a-thread-width wide.
    // I've forgotten what the +1 is for, probably to account for
    // the separating line

    rx = (location.x * spacing) + (fabricWidth * location.x) 
 	  + fabricWidth / 2.0 + 1;
    ry = (location.y * spacing) + (fabricWidth * location.y) 
	  + fabricWidth / 2.0 + 1;

    // make the actual cell rect
    NSRect rect;
    rect = NSMakeRect (rx, ry, spacing, spacing);

    return (rect);

} // rectForLocation



// given an x and y in view coordinates, indicate which cell the
// point is over.  (-1, -1) is returned for no cell under the mouse
// (for things like mouse over the grid lines

- (void) mapPoint: (NSPoint) point  toX: (int *) x  Y: (int *) y
{
    // this is pretty much the inverse operation to rectForLocation

    float fabricWidth, spacing;
    fabricWidth = [group fabricWidth];
    spacing = [group fabricSpacing];

    *x = (point.x - fabricWidth / 2.0) / (fabricWidth + spacing);
    *y = (point.y - fabricWidth / 2.0) / (fabricWidth + spacing);

    // make sure we're actually in the cell
    BWStitchLocation loc;
    loc = BWMakeStitchLocation (*x, *y);

    NSRect rect;
    rect = [self rectForLocation: loc];

    if (!NSPointInRect (point, rect)) {
	*x = -1;
	*y = -1;
    }

} // mapPoint



// for the woven-fabric effect, first draw a big gray grid, then
// draw the shorter black lines to give the illusion of the weave

- (void) drawGrid: (NSRect) rect
{
    [NSGraphicsContext saveGraphicsState]; {

	// set the width of the pen to match the fabric width
	float fabricWidth;
	fabricWidth = [group fabricWidth];
	[NSBezierPath setDefaultLineWidth: fabricWidth];

	// get the color specified in the model
	[[group fabricColor] set];

	// this is how far to skip between the grid lines
	float stride;
	stride = fabricWidth + [group fabricSpacing];

	// draw vertical lines
	float x;
	for (x = rect.origin.x; x <= rect.origin.x + rect.size.width; 
	     x += stride) {
	    NSPoint start, end;
	    start = NSMakePoint (x, rect.origin.y);
	    end = NSMakePoint (x, rect.origin.y + rect.size.height);
	    [NSBezierPath strokeLineFromPoint: start  toPoint: end];
	}
	
	// draw the horizontal lines
	float y;
	for (y = rect.origin.y; y <= rect.origin.y + rect.size.height; 
	     y += stride) {
	    NSPoint start, end;
	    start = NSMakePoint (rect.origin.x, y);
	    end = NSMakePoint (rect.origin.x + rect.size.width, y);
	    [NSBezierPath strokeLineFromPoint: start  toPoint: end];
	}
	
    } [NSGraphicsContext restoreGraphicsState];

} // drawGrid



// draw the lines to give the illusion of woven fabric

- (void) drawOverlapLines: (NSRect) rect
{
    [NSGraphicsContext saveGraphicsState]; {
	
	[NSBezierPath setDefaultLineWidth: 1.0];
	[[NSColor blackColor] set];

	float spacing, fabricWidth;
	spacing = [group fabricSpacing];
	fabricWidth = [group fabricWidth];
	
	// draw two cell withds (plus grid line), skip one grid line, ...
	float twoWidth;
	twoWidth = spacing * 2 + fabricWidth;
	
	// this is the amount of space for the 'skip one'
	float stride;
	stride = fabricWidth + spacing;

	float vstride;
	vstride = twoWidth + fabricWidth;

	float bias = fabricWidth / 2.0;

	// change the start for every other line
	BOOL jiggle = NO;
	
	float x, y, ystart;
	for (x = rect.origin.x - bias; x < rect.origin.x + rect.size.width; x += stride) {
	    if (jiggle) {
		ystart = rect.origin.y;
	    } else {
		// start off the top
		ystart = rect.origin.y - spacing - fabricWidth;
	    }
	    jiggle = !jiggle;
	    
	    for (y = ystart + bias; y < rect.origin.y + rect.size.height;
		 y += vstride) {
		NSPoint start, end;
		
		start = NSMakePoint (x, y);
		end = NSMakePoint (x, y + twoWidth);
		[NSBezierPath strokeLineFromPoint: start  toPoint: end];
		
		start = NSMakePoint (x + fabricWidth, y);
		end = NSMakePoint (x + fabricWidth, y + twoWidth);
		[NSBezierPath strokeLineFromPoint: start  toPoint: end];
	    }
	}
	
	jiggle = YES;
	
	float xstart;
	for (y = rect.origin.y - bias; y < rect.origin.y + rect.size.height; y += stride) {
	    if (jiggle) {
		xstart = rect.origin.x;
	    } else {
		// start off the left
		xstart = rect.origin.x - spacing - fabricWidth;
	    }
	    jiggle = !jiggle;
	    
	    for (x = xstart + bias; x < rect.origin.x + rect.size.width;
		 x += vstride) {
		NSPoint start, end;
		
		start = NSMakePoint (x, y);
		end = NSMakePoint (x + twoWidth, y);
		[NSBezierPath strokeLineFromPoint: start  toPoint: end];
		
		start = NSMakePoint (x, y + fabricWidth);
		end = NSMakePoint (x + twoWidth, y + fabricWidth);
		[NSBezierPath strokeLineFromPoint: start  toPoint: end];
	    }
	}
    } [NSGraphicsContext restoreGraphicsState];

} // drawOverlapLines



// standard distance formula stuff

- (float) distanceFromPoint: (NSPoint) point1
		    toPoint: (NSPoint) point2
{
    float x2mx1, y2my1;
    x2mx1 = point2.x - point1.x;
    y2my1 = point2.y - point1.y;

    float length;
    length = sqrt ((x2mx1 * x2mx1) + (y2my1 * y2my1));

    return (length);

} // distanceFromPoint1



// given two points, calculate the angle (in radians) the line formed
// by the two points, and a horizontal line.

- (float) calculateAngleFromPoint: (NSPoint) point1  toPoint: (NSPoint) point2
{
    float hyptLength;
    hyptLength = [self distanceFromPoint: point1  toPoint: point2];

    // normalize the vector

    // move the vector to the origin
    point2.x -= point1.x;
    point2.y -= point1.y;

    // scale it
    point2.x /= hyptLength;
    point2.y /= hyptLength;

    float angle;
    angle = atan2 (point2.y, point2.x);

    return (angle);

} // calculateAngleFromPoint



// generate an oblong oval for the stitch the user has placed
// on the canvas.  This makes a vertical shape, which is then
// later drawn into a rotated coordinate system.  The two points
// are just sued for determining the 

- (NSBezierPath *) makeStitchWithLength: (float) length
				  width: (float) width
{
    float angle1, angle2;

    // these are the start and stop angles for the half-circle
    // end points
    angle1 = 90.0;
    angle2 = 270.0;

    // calcuate the four points around the rectangle
    NSPoint pt1, pt2;
    NSPoint pt3, pt4;

    pt1 = NSMakePoint (0.0, width);
    pt2 = NSMakePoint (0.0, -width);
    
    pt3 = NSMakePoint (length, -width);
    pt4 = NSMakePoint (length, width);

    // center point of the far-end
    NSPoint endCenter;
    endCenter = NSMakePoint (length, 0.0);

    // make the new path
    NSBezierPath *path;
    path = [NSBezierPath bezierPath];

    // start at one corner of the rectangle
    [path moveToPoint: pt1];
    
    // add the top half-circle
    [path appendBezierPathWithArcWithCenter: NSZeroPoint
	  radius: width
	  startAngle: angle1
	  endAngle: angle2];

    // move to the bottom part of the rectangle
    [path lineToPoint: pt3];

    // add the bottom half-circle
    [path appendBezierPathWithArcWithCenter: endCenter
	  radius: width
	  startAngle: angle2
	  endAngle: angle1];

    // complete the path
    [path lineToPoint: pt1];
    
    return (path);

} // makeStitchWithLength



// draw the stitch (the ovalThing, you make my heart sing)

- (void) drawOvalthingFromPoint: (NSPoint) point1
			toPoint: (NSPoint) point2
			  width: (float) width
{
    [NSGraphicsContext saveGraphicsState]; {

	// make the oval path.  This is in a vertical orientation,
	// centered at the origin

	// how long to make the path
	float length;
	length = [self distanceFromPoint: point1
		       toPoint: point2];

	NSBezierPath *path;
	path = [NSBezierPath bezierPath];
	path = [self makeStitchWithLength: length
		     width: width];

	// time to mutate the coordinate system so it'll put
	// the path in the proper place
	NSAffineTransform *transform;
	transform = [NSAffineTransform transform];

	// move the origin to the start point
	[transform translateXBy: point1.x  yBy: point1.y];

	// rotate to the angle specified by the two points
	float angle;
	angle = [self calculateAngleFromPoint: point1  toPoint: point2];

	[transform rotateByRadians: angle];

	// now all subseuqent drawing will be affected by the
	// transform matrix
	[transform concat];

	[path setLineWidth: 1.5];

	// fill in with the stitch color
	[[group stitchColor] set];
	[path fill];

	// stroke an outline
	[[NSColor blackColor] set];
	[path stroke];

    } [NSGraphicsContext restoreGraphicsState];

} // drawOvalthingFromPoint



// calculate point relative from center where the label should go
// there are 8 positions around the circle to put the number:
// (pretend these are uniform size angles, each 45 degrees
//    4  3  2
//     \ | /
//      \|/
// 5 ----+---- 1
//      /|\
//     / | \
//    6  7  8
// straddling each position is a 45 degree angle (offset by 22.5 degrees,
// or half the 45 degree angle.  Those are used to figure out the
// quadrants

typedef enum {
    E = 1, NE, N, NW, W, SW, S, SE
} labelPosition;



// angle is in degrees

- (labelPosition) getOctantForAngle: (float) angle
{
    labelPosition result = E;

    if (angle > 337.5  || angle <= 22.5) {
	result = E;
    } else if (angle > 22.5 && angle <= 67.5) {
	result = NE;
    } else if (angle > 67.5 && angle <= 112.5) {
	result = N;
    } else if (angle > 112.5 && angle <= 157.5) {
	result = NW;
    } else if (angle > 157.5 && angle <= 202.5) {
	result = W;
    } else if (angle > 202.5 && angle <= 247.5) {
	result = SW;
    } else if (angle > 247.5 && angle <= 292.5) {
	result = S;
    } else if (angle > 292.5 && angle <= 337.5) {
	result = SE;
    } else {
	// bad angle, which would mean that the start and
	// stop are in the same square.
    }

    return (result);

} // getPointForAngle



// given a point, offset it based on the octant specified above.
// This is where the numeric label will be placed.  That way it's
// directly off the end of the stitch

- (NSPoint) offsetPoint: (NSPoint) point  inPosition: (labelPosition) position
{
    // these numbers were tweaked until it looked goo
    float offset;
    offset = 3.0;

    float horizontalBias;
    horizontalBias= 4.0;

    switch (position) {
    case E:
	point.x += offset + horizontalBias;
	break;
    case NE:
	point.x += offset;
	point.y -= offset;
	break;
    case N:
	point.y -= offset;
	break;
    case NW:
	point.x -= offset;
	point.y -= offset;
	break;
    case W:
	point.x -= offset + horizontalBias;
	break;
    case SW:
	point.x -= offset;
	point.y += offset;
	break;
    case S:
	point.y += offset;
	break;
    case SE:
	point.x += offset;
	point.y += offset;
	break;
    }

    return (point);

} // offsetPoint



// given a label string, and a position to put the label,
// figure out what rectangle to use to draw it in.  

- (NSRect) calculateRectForString: (NSString *) string
		       atPosition: (labelPosition) position
{
    // calculate the size of the string
    NSSize stringSize;
    stringSize = [string sizeWithAttributes: textAttributes];

    NSPoint rectOrigin;

    // figure out where to orient the rectangle
    switch (position) {
    case E:
	rectOrigin.x = 0.0;
	rectOrigin.y = - (stringSize.height / 2);
	break;
    case NE:
	rectOrigin.x = 0.0;
	rectOrigin.y = - stringSize.height;
	break;
    case N:
	rectOrigin.x = - (stringSize.width / 2);
	rectOrigin.y = - stringSize.height;
	break;
    case NW:
	rectOrigin.x = - stringSize.width;
	rectOrigin.y = - stringSize.height;
	break;
    case W:
	rectOrigin.x = - stringSize.width;
	rectOrigin.y = - (stringSize.height / 2);
	break;
    case SW:
	rectOrigin.x = -stringSize.width;
	rectOrigin.y = 0.0;
	break;
    case S:
	rectOrigin.x = - (stringSize.width / 2);
	rectOrigin.y = 0.0;
	break;
    case SE:
	rectOrigin.x = 0.0;
	rectOrigin.y = 0.0;
	break;
    }

    NSRect stringRect;
    stringRect.origin = rectOrigin;
    stringRect.size = stringSize;

    return (stringRect);

} // calculateRectForString



// given two numbers for labels, and the position where the stitch
// starts and ends, display the numbers

- (void) drawNumbersForStart: (int) start  end: (int) end
		     center1: (NSPoint) center1  center2: (NSPoint) center2
{
    BOOL displayStart = NO, displayEnd = NO;

    // figure out if we need to draw a particular class of numbers
    switch ([group numberDisplay]) {

    case BWStitchGroup_EvenOdd:
	displayStart = YES;
	displayEnd = YES;
	break;

    case BWStitchGroup_Even:
	displayStart = !(start % 2);
	displayEnd = !(end % 2);
	break;

    case BWStitchGroup_Odd:
	displayStart = start % 2;
	displayEnd = end % 2;
	break;

    case BWStitchGroup_None:
	displayStart = NO;
	displayEnd = NO;
	break;
    }

    [NSGraphicsContext saveGraphicsState]; {

	// go from zero base to 1 base, since these numbers
	// will be for human consumption (for internal use only)
	start++;
	end++;

	// figure out the angle, which determines where the label will
	// be placed relative to the stitch shape
	float angle;
	angle = [self calculateAngleFromPoint: center1  toPoint: center2];
	
	// the angle is 0 to pi for quadrant IV to III
	// and 0 to -pi for quadrant I to II
	
	// convert from radians to degrees
	float degrees;

	if (angle <= 0.0) {
	    degrees = 180 * fabs(angle) / 3.14159;
	} else {
	    degrees = 360 - 180 * fabs(angle) / 3.14159;
	}
	
	// figure out what position to put the label
	labelPosition position;
	position = [self getOctantForAngle: degrees];

	// figure out where the offset goes relative to the center
	// of the far end of the stitch
	NSPoint point;
	point = [self offsetPoint: center2  inPosition: position];

	// convert the number into a string, which we can then
	// draw with
	NSString *string;
	string = [NSString stringWithFormat: @"%d", end];

	// figure out exactly the rectangle to stash the string
	NSRect stringRect;
	stringRect = [self calculateRectForString: string
			   atPosition: position];

	// offset the string rect from the starting point
	stringRect.origin.x += point.x;
	stringRect.origin.y += point.y;

	// draw it, if the user hasn't told us not to
	if (displayStart) {
	    [string drawInRect: stringRect  withAttributes: textAttributes];
	}
	
	// move to the other side of the circle and do the same thing
	
	position = (((position - 1) + 4) % 8) + 1;

	point = [self offsetPoint: center1
		      inPosition: position];
	
	string = [NSString stringWithFormat: @"%d", start];
	stringRect = [self calculateRectForString: string
			   atPosition: position];
	
	stringRect.origin.x += point.x;
	stringRect.origin.y += point.y;
	
	if (displayEnd) {
	    [string drawInRect: stringRect  withAttributes: textAttributes];
	}
	
    } [NSGraphicsContext restoreGraphicsState];

} // drawNumbersForStart



// draw all the stitches the user has laid down

- (void) drawStitches: (NSRect) rect
{
    [NSGraphicsContext saveGraphicsState]; {

	// number of locations we'll be doing
	int count;
	count = [group count];

	// is there an odd number of locations?  If so draw stitch
	// for the pairs of locations.  If there's one left over,
	// it'll be drawn differently
	int odd;
	odd = (count % 2);
	
	// how many locations to process drawing oval stitches?
	int numToProcess = count;
	if (odd) {
	    numToProcess--;
	}

	// walk through pairs of locations
	int i;
	for (i = 0; i < numToProcess; i += 2) {

	    // get the locations of either end
	    BWStitchLocation loc1, loc2;
	    loc1 = [group stitchLocationAtIndex: i];
	    loc2 = [group stitchLocationAtIndex: i + 1];

	    // get the cell rects for either end.  The stitches
	    // have their control points in the center of these
	    // rectangles
	    NSRect rect1, rect2;
	    rect1 = [self rectForLocation: loc1];
	    rect2 = [self rectForLocation: loc2];
	    
	    // connect the centers
	    NSPoint center1, center2;

	    center1 = NSMakePoint (rect1.origin.x + rect1.size.width / 2.0,
				   rect1.origin.y + rect1.size.height / 2.0);
	    center2 = NSMakePoint (rect2.origin.x + rect2.size.width / 2.0,
				   rect2.origin.y + rect2.size.height / 2.0);

	    // draw the stitch
	    [self drawOvalthingFromPoint: center1  
		  toPoint: center2  
		  width: [group stitchSize]];

	    // draw the labels (if appropriate)
	    [self drawNumbersForStart: i
		  end: i + 1
		  center1: center1
		  center2: center2];
	}

	
	// draw a dot for the odd one out
	if (odd) {
	    // get the last location
	    BWStitchLocation loc;
	    loc = [group stitchLocationAtIndex: count - 1];

	    // find the rect for that cell
	    NSRect rect;
	    rect = [self rectForLocation: loc];
	    
	    // make the dot a little smaller than the cell
	    rect = NSInsetRect (rect, 2.0, 2.0);

	    // draw it
	    NSBezierPath *path;
	    path = [NSBezierPath bezierPathWithOvalInRect: rect];
	    
	    [[NSColor blueColor] set];
	    [path fill];
	}
    } [NSGraphicsContext restoreGraphicsState];
	
} // drawStitches



// draw the highlighting to indicate the printable area

- (void) drawPrintableArea: (NSRect) rect
{
    BWIntRect printableArea;

    if (trackingPrintArea) {
	// if we're tracking, use the fake rect since it'll
	// be "more current" that what's in the model
	printableArea = BWMakeIntRect (printAreaX, printAreaY,
				       printAreaW, printAreaH);
    } else {
	printableArea = [group printableArea];
    }

    // get the top-left corner.
    BWStitchLocation topLeft;
    topLeft = BWMakeStitchLocation (printableArea.origin.x,
				    printableArea.origin.y);

    if (printableArea.size.width == 0 || printableArea.size.height == 0) {
	// It's an empty rectangle.

	// if we're tracking and we've already picked one corner, 
	// draw something there to give some feedback

	if (trackingPrintArea && gotTopLeft) {
	    NSRect cornerRect;
	    cornerRect = [self rectForLocation: topLeft];
	    // offset into fabric
	    cornerRect.origin.x -= [group fabricWidth] / 2.0;
	    cornerRect.origin.y -= [group fabricWidth] / 2.0;
	    cornerRect.size.width = [group fabricWidth];
	    cornerRect.size.height = [group fabricWidth];

	    [NSGraphicsContext saveGraphicsState]; {
		[[NSColor purpleColor] set];
		[NSBezierPath setDefaultLineWidth: 3];
		[NSBezierPath strokeRect: cornerRect];
	    } [NSGraphicsContext restoreGraphicsState];
	}
	
	// don't worry about drawing the rest of the rectangle
	goto bailout;
    }

    // calculte the rectangle.  Find the real rectangles for
    // the two and union them
    NSRect topLeftStitchRect;
    topLeftStitchRect = [self rectForLocation: topLeft];

    // get the bottom-right corner.
    BWStitchLocation bottomRight;
    bottomRight = BWMakeStitchLocation (printableArea.origin.x 
					+ printableArea.size.width,
					printableArea.origin.y 
					+ printableArea.size.height);

    NSRect bottomRightStitchRect;
    bottomRightStitchRect = [self rectForLocation: bottomRight];

    // union them
    NSRect unionRect;
    unionRect = NSUnionRect (topLeftStitchRect, bottomRightStitchRect);

    // the fabricWidth / 2.0 to run the printable area between the
    // fabric threads
    unionRect = NSInsetRect (unionRect, 
			     -[group fabricWidth] / 2.0,
			     -[group fabricWidth] / 2.0);

    // stroke the resulting area
    [NSGraphicsContext saveGraphicsState]; {
	[[NSColor purpleColor] set];
	[NSBezierPath setDefaultLineWidth: 3];
	[NSBezierPath strokeRect: unionRect];
    } [NSGraphicsContext restoreGraphicsState];

bailout:
    return;

} // drawPrintableArea



// draw the mouse motion tracking, which is just darkening in
// the cell the mouse is over

- (void) drawTracking
{
    if (trackX != BWFabricView_NO_TRACK
	|| trackY != BWFabricView_NO_TRACK) {

	// trackX and trackY get set in mouseMoved
	BWStitchLocation loc;
	loc = BWMakeStitchLocation (trackX, trackY);
	
	NSRect cellRect;
	cellRect = [self rectForLocation: loc];

	[NSGraphicsContext saveGraphicsState]; {
	    [cellHighlightColor set];
	    [NSBezierPath fillRect: cellRect];
	} [NSGraphicsContext restoreGraphicsState];
    }

} // drawTracking



// draw the view. 

- (void) drawRect: (NSRect) rect
{
    rect = [self bounds];

    // fill in the whole background with the appropriate color
    [[group backgroundColor] set];
    [NSBezierPath fillRect: rect];

    // inset a bit to skip over the border around the whole thing
    NSRect innerRect;
    innerRect = NSInsetRect (rect, 1.0, 1.0);

    // draw the individual parts
    [self drawGrid: innerRect];
    [self drawOverlapLines: innerRect];

    // don't draw the tracking when printing
    if ([NSGraphicsContext currentContextDrawingToScreen]) {
	[self drawTracking];
    }

    [self drawStitches: innerRect];

    // don't draw printable area when printing
    if ([NSGraphicsContext currentContextDrawingToScreen]) {
	[self drawPrintableArea: innerRect];
    }

    [NSBezierPath setDefaultLineWidth: 4.0];

    // draw a nice border around everything
    [[NSColor blackColor] set];
    [NSBezierPath strokeRect: rect];

} // drawRect



// we're flipping out!  wooooo hoooo!!!!
// It makes a lot of the math easier to use an upper-left coordinate
// system

- (BOOL) isFlipped
{
    return (YES);
} // isFlipped



// the user-interface is a little odd, since we don't track out the
// stitch live (since the original design idea was to have the
// individual stitch locations being undoable.  So there's not really
// anything to do with mouse down and mouse drag.  MouseUp's are where
// the interesting stuff happens

- (void) mouseUp: (NSEvent *) event
{
    // where's the click?
    NSPoint mouse;
    mouse = [self convertPoint: [event locationInWindow]
		  fromView: nil];

    // there are two cases - are we tracking the print area,
    // or is it a random click.

    if (trackingPrintArea) {
	int x, y;

	// we want the printable area to go between the grid lines,
	// so offset the mouse point so that it actually falls
	// in a cell for doing hit-testing

	if (!gotTopLeft) {
	    // offset so that the mouse is in the cell rect, but
	    // appears to be in the stitch area
	    mouse.x += [group fabricWidth];
	    mouse.y += [group fabricWidth];
	} else {
	    mouse.x -= [group fabricWidth];
	    mouse.y -= [group fabricWidth];
	}

	// figure out the cell associated with the point
	[self mapPoint: mouse
	      toX: &x
	      Y: &y];

	// not a click in a useful area
	if (x == -1 || y == -1) {
	    return;
	}

	if (gotTopLeft) {
	    // we've gotten the top-left corner already, so
	    // this is for the bottom-right
	    printAreaW = x - printAreaX;
	    printAreaH = y - printAreaY;

	    [self stopTrackingPrintableArea];

	    NSNotificationCenter *center;
	    center = [NSNotificationCenter defaultCenter];

	    // tell observers there's a new printable area.  They
	    // should then update the model
	    [center postNotificationName: BWFabricViewPrintableAreaNotification
		    object: self
		    userInfo: nil];
	} else {
	    // this is the first click, so it's the top-left
	    printAreaX = x;
	    printAreaY = y;

	    gotTopLeft = YES;

	    // change cursors
	    [[self window] invalidateCursorRectsForView: self];
	}
	[self setNeedsDisplay: YES];

    } else {
	// this is the add a stitch location case
	int x, y;

	// where are we?
	[self mapPoint: mouse
	      toX: &x
	      Y: &y];

	// we're in a cell.  Yay
	if (x != -1 && y != -1) {
	    // which cell?
	    BWStitchLocation loc;
	    loc = BWMakeStitchLocation (x, y);

	    // broadcast the click.  The observer will add the location
	    // to the model
	    NSValue *location;
	    location = [NSValue valueWithBytes: &loc
				objCType: @encode(BWStitchLocation)];
	
	    NSNotificationCenter *center;
	    center = [NSNotificationCenter defaultCenter];

	    // just pass the location, rather than wrapping it into
	    // a dictionary.  Lazy, yeah.
	    [center postNotificationName: BWFabricViewNewStitchNotification
		    object: self
		    userInfo: (NSDictionary*)location];    // hack

	    [self setNeedsDisplay: YES];
	}
    }

} // mouseUp



// handle mouse-moved events.  Highlight the cell underneath the mouse
// (if it is over a cell)

- (void) mouseMoved: (NSEvent *) event
{
    NSPoint mouse;
    mouse = [self convertPoint: [event locationInWindow]
		  fromView: nil];
    
    if (trackingPrintArea) {
	// don't do anything since tracking the print area is
	// a modal operation

    } else {
	int x, y;

	// where are we?
	[self mapPoint: mouse
	      toX: &x
	      Y: &y];

	// only update the location if it changed.  No need to re-draw
	// the same thing all the time
	if (x != trackX || y != trackY) {

	    BWStitchLocation xyloc;
	    xyloc = BWMakeStitchLocation (trackX, trackY);

	    // inval the old
	    NSRect rect;
	    rect = [self rectForLocation: xyloc];
	    [self setNeedsDisplayInRect: rect];

	    trackX = x;
	    trackY = y;
	    
	    // inval the new
	    xyloc = BWMakeStitchLocation (trackX, trackY);
	    rect = [self rectForLocation: xyloc];
	    [self setNeedsDisplayInRect: rect];
	}
    }
} // mouseMoved



// this is a notification from the model, where someone has changed
// some change that affects the display of the view.

- (void) groupVisualChange: (NSNotification *) notification
{
    // for text stuff, we update our attributes dict
    if ([group textFont]) {
	[textAttributes setObject: [group textFont]
			forKey: NSFontAttributeName];
    }

    if ([group textColor]) {
	[textAttributes setObject: [group textColor]
			forKey: NSForegroundColorAttributeName];
    }

    // in all cases, force a redraw to use the new attribute

    [self setNeedsDisplay: YES];

} // groupVisualChange



// associate the model with us, the view

- (void) setStitchGroup: (BWStitchGroup *) newGroup
{
    [newGroup retain];

    // in with the new
    // out with the old
    [group release];
    group = newGroup;

    // register for updates
    NSNotificationCenter *center;
    center = [NSNotificationCenter defaultCenter];

    [center addObserver: self
	    selector: @selector(groupVisualChange:)
	    name: BWStitchGroup_VisualAttributeChangeNotification
	    object: group];

    // make text attributes with the font and color

    [textAttributes release];
    textAttributes = [[NSMutableDictionary alloc] init];
    
    if ([group textFont]) {
	[textAttributes setObject: [group textFont]
			forKey: NSFontAttributeName];
    }
    if ([group textColor]) {
	[textAttributes setObject: [group textColor]
			forKey: NSForegroundColorAttributeName];
    }

    // we've got a new model, so a totally new display
    [self setNeedsDisplay: YES];

} // setStitchGroup



// someone has selected an action for making a new printable area

- (void) startTrackingPrintableArea
{
    // zero out the printable area
    printAreaX = printAreaY = printAreaW = printAreaH = 0;

    // set mode flags
    trackingPrintArea = YES;
    gotTopLeft = NO;
    
    // force a cursor change to an upper-left bracket
    [[self window] invalidateCursorRectsForView: self];

} // startTrackingPrintableArea



// ok, we're done tracking

- (void) stopTrackingPrintableArea
{
    // clear mode flags
    trackingPrintArea = NO;
    gotTopLeft = NO;

    // force a cursor change back to the i-beam
    [[self window] invalidateCursorRectsForView: self];

} // stopTrackingPrintableArea



// get the print area that the user selected

- (BWIntRect) getTrackingPrintArea
{
    BWIntRect rect;
    rect = BWMakeIntRect (printAreaX, printAreaY,
			  printAreaW, printAreaH);

    return (rect);

} // getTrackingPrintArea



// calculate the printable size, the width and height of the
// printable area
- (NSSize) printableSize
{
    BWIntRect printableArea;
    printableArea = [group printableArea];
    
    // empty rect.  try to figure out the right thing to print
    if (printableArea.size.width == 0 || printableArea.size.height == 0) {
	printableArea = [group contentBounds];
	printableArea = [self tweakPrintableArea: printableArea];
    }
    

    // +2 so that the right size has enough space for the half-thread that
    // straddles the border
    NSSize printableSize;
    printableSize.width 
	= (printableArea.size.width + 1)
	  * ([group fabricWidth] + [group fabricSpacing]) + 2.0;

    printableSize.height
	= (printableArea.size.height + 1)
	  * ([group fabricWidth] + [group fabricSpacing]) + 2.0;
    
    return (printableSize);

} // printableSize



// we need to do this to get mouse-moved events

- (BOOL) acceptsFirstResponder
{
    // for mouse-moved
    return (YES);

} // acceptsFirstResponder



// the view changed such that the cursor rects need to be updated

- (void) resetCursorRects
{
    if (trackingPrintArea) {

	if (!gotTopLeft) {
	    [self addCursorRect: [self bounds]
		  cursor: [BWCursor upperLeftCursor]];
	} else {
	    [self addCursorRect: [self bounds]
		  cursor: [BWCursor bottomRightCursor]];
	}

    } else {
	[self addCursorRect: [self bounds]
	      cursor: [BWCursor crosshairCursor]];
    }

} // resetCursorRects


@end // BWFabricView

