// BWStitchGroup.h -- the model object for the stitch notator.  It
//                    contains everything that would need to be
//                    saved to a file.

#import <Cocoa/Cocoa.h>

// notification that is broadcast when a visual attribute, like
// thread size or stitch color is changed, so that any views will
// automatically update themselves

extern NSString *BWStitchGroup_VisualAttributeChangeNotification;



// Kind of like an NSPoint, but using integers.  Stitch locations
// have a top-left origin of 0,0.
// The stitchGroup keeps a sequence of stitch locations.  Each pair
// of locations defines a stitch.

typedef struct BWStitchLocation {
    int x, y;
} BWStitchLocation;


// the controls for which numbers to display when drawing a stitch group.

typedef enum {
    BWStitchGroup_EvenOdd,
    BWStitchGroup_Even,
    BWStitchGroup_Odd,
    BWStitchGroup_None
} BWStitchGroup_NumberDisplay;


// control constants for shifing all the stitches around in the
// group.  Useful if the user started too close to the edge of the view

typedef enum {
    BWStitchGroup_ShiftUp,
    BWStitchGroup_ShiftDown,
    BWStitchGroup_ShiftLeft,
    BWStitchGroup_ShiftRight   // pop up push down byte byte byte!
} BWStitchGroup_ShiftDirection;



// like an NSRect, but just integers.  Represents ranges of stitch
// locations

typedef struct BWIntRect {
    struct origin {
	int x;
	int y;
    } origin;
    struct size {
	int width;
	int height;
    } size;
} BWIntRect;



// and finally, we come to the actual class
@interface BWStitchGroup : NSObject <NSCoding>
{
    // storage for the stitches.  Just an array of locations.
    // the actual number of stitches will be small (less than several
    // dozen), so no need for a fancier data structure

    int pointCount;
    BWStitchLocation *points;

    // display-related values

    float threadSize;
    float spacing;
    float stitchSize;

    // even only, odd only, even/odd, none
    BWStitchGroup_NumberDisplay numberDisplay;

    NSColor *fabricColor;
    NSColor *stitchColor;
    NSColor *backgroundColor;
    NSColor *textColor;

    NSFont *textFont;

    // this is the area that the user will want printed.
    // if it's empty, the contentBounds function can be used
    // to guess a good printableArea
    BWIntRect printableArea;
}

- (id) init;

- (int) count;
- (void) addStitchLocation: (BWStitchLocation) stitch;
- (BWStitchLocation) stitchLocationAtIndex: (int) index;

// undo is going to be the only way to remove stitches,
// so just removing the last one is suffficient functionality
- (void) removeLastStitchLocation;

- (void) shiftStitchesBy: (int) number  
	       direction: (BWStitchGroup_ShiftDirection) direction;

- (void) setFabricWidth: (float) width;
- (float) fabricWidth;

- (void) setFabricSpacing: (float) spacing;
- (float) fabricSpacing;

- (void) setStitchSize: (float) stitchSize;
- (float) stitchSize;

- (void) setFabricColor: (NSColor *) color;
- (NSColor *) fabricColor;

- (void) setStitchColor: (NSColor *) color;
- (NSColor *) stitchColor;

- (void) setBackgroundColor: (NSColor *) color;
- (NSColor *) backgroundColor;

- (void) setTextColor: (NSColor *) color;
- (NSColor *) textColor;

- (void) setTextFont: (NSFont *) font;
- (NSFont *) textFont;

- (void) setNumberDisplay: (BWStitchGroup_NumberDisplay) display;
- (BWStitchGroup_NumberDisplay) numberDisplay;

- (void) setPrintableArea: (BWIntRect) printableArea;
- (void) clearPrintableArea;
- (BWIntRect) printableArea;

// calculate a printable area to hold all the stitches
- (BWIntRect) contentBounds;


@end // BWStitchGroup



// some convenience in-lines to create new intrects and
// stitch locations

FOUNDATION_STATIC_INLINE 
BWIntRect BWMakeIntRect (int x, int y, int w, int h) 
{
    BWIntRect rect;
    rect.origin.x = x;
    rect.origin.y = y;
    rect.size.width = w;
    rect.size.height = h;

    return (rect);

} // BWMakeIntRect


FOUNDATION_STATIC_INLINE
BWStitchLocation BWMakeStitchLocation (int x, int y)
{
    BWStitchLocation location;

    location.x = x;
    location.y = y;

    return (location);

} // BWMakeStitchLocation
