// BWStitchGroup.m -- the model object for the stitch notator.  It
//                    contains everything that would need to be
//                    saved to a file.

#import "BWStitchGroup.h"


// notification that is broadcast when a visual attribute, like
// thread size or stitch color is changed, so that any views will
// automatically update themselves

NSString *BWStitchGroup_VisualAttributeChangeNotification
    = @"BWStitchGroup Visual Attribute Change Notification";


@implementation BWStitchGroup

// given a skull full of mush, turn it into a proper stitch group

- (id) init
{
    if (self = [super init]) {

	// storage for the stitches
	points = malloc (sizeof(BWStitchLocation));

	// defaults, chosen arbitrarily, but they tend to look
	// pretty good
	threadSize = 9.0;
	spacing = 20.0;
	stitchSize = 5.0;

	// make a bold font to be used for the labels
	textFont = [NSFont fontWithName: @"Helvetica"  
			   size: 18];
    
	NSFontManager *fontManager;
	fontManager = [NSFontManager sharedFontManager];

	textFont = [fontManager convertFont: textFont
				toHaveTrait: NSBoldFontMask];

	// this is the color that is used for the fabric grid.
	// the default lightGrayColor is just a skootch too 
	// dark
	fabricColor = [NSColor colorWithCalibratedRed: 0.78
			       green: 0.78
			       blue: 0.78
			       alpha: 1.0];
	[fabricColor retain];

	// draw stitches in white

	stitchColor = [NSColor whiteColor];
	// yeah, whiteColor and blackColor are singletons and will
	// live for the life of the program, but go ahead and
	// do the retain/release work since we'll be needing to
	// release all our colors in dealloc.
	[stitchColor retain];

	// the background is a very light gray, just dark enough
	// so that white stitches stand out a little bit
	backgroundColor = [NSColor colorWithCalibratedRed: 0.95
				   green: 0.95
				   blue: 0.95
				   alpha: 1.0];
	[backgroundColor retain];

	textColor = [NSColor blackColor];
	[textColor retain];

	// default to an empty printable area
	printableArea = BWMakeIntRect (0, 0, 0, 0);
    }

    return (self);

} // init


// save ourselves to the coder stream.  Like when saving a document

- (void) encodeWithCoder: (NSCoder *) aCoder
{
    // first write out the version of the document
    int version = 1;
    [aCoder encodeValueOfObjCType: @encode(int)
	    at: &version];

    // pretty much just chug through the class structure
    [aCoder encodeValueOfObjCType: @encode(int)
	    at: &pointCount];

    int i;
    for (i = 0; i < pointCount; i++) {
	[aCoder encodeValueOfObjCType: @encode(BWStitchLocation)
		at: &points[i]];
    }
    
    [aCoder encodeValueOfObjCType: @encode(float)
	    at: &threadSize];
    [aCoder encodeValueOfObjCType: @encode(float)
	    at: &spacing];
    [aCoder encodeValueOfObjCType: @encode(float)
	    at: &stitchSize];

    [aCoder encodeValueOfObjCType: @encode(BWStitchGroup_NumberDisplay)
	    at: &numberDisplay];

    [aCoder encodeObject: fabricColor];
    [aCoder encodeObject: stitchColor];
    [aCoder encodeObject: backgroundColor];
    [aCoder encodeObject: textColor];

    [aCoder encodeObject: textFont];

    [aCoder encodeValueOfObjCType: @encode(BWIntRect)
	    at: &printableArea];

} // encodeWithCoder



// create a new group from a coder, like when reading in a document

- (id) initWithCoder: (NSCoder *) aDecoder
{
    if (self = [super init]) {

	// decode stuff in the same order as the encoder
	
	// get the version
	int version;
	[aDecoder decodeValueOfObjCType: @encode(int)
		  at: &version];

	if (version != 1) {
	    // really should put up some error to the user
	    NSLog (@"Version is too new.  Can't read it.  Sorry");
	    self = nil;
	    goto cannot_load;
	}

	[aDecoder decodeValueOfObjCType: @encode(int)
		  at: &pointCount];

	points = realloc (points, sizeof(BWStitchLocation) * pointCount);
	int i;
	for (i = 0; i < pointCount; i++) {
	    [aDecoder decodeValueOfObjCType: @encode(BWStitchLocation)
		      at: &points[i]];
	}

	[aDecoder decodeValueOfObjCType: @encode(float)
		  at: &threadSize];
	[aDecoder decodeValueOfObjCType: @encode(float)
		  at: &spacing];
	[aDecoder decodeValueOfObjCType: @encode(float)
		  at: &stitchSize];

	[aDecoder decodeValueOfObjCType: @encode(BWStitchGroup_NumberDisplay)
		  at: &numberDisplay];

	fabricColor = [[aDecoder decodeObject] retain];
	stitchColor = [[aDecoder decodeObject] retain];
	backgroundColor = [[aDecoder decodeObject] retain];
	textColor = [[aDecoder decodeObject] retain];

	textFont = [[aDecoder decodeObject] retain];

	[aDecoder decodeValueOfObjCType: @encode(BWIntRect)
		  at: &printableArea];
    }

cannot_load:
    return (self);

} // initWithCoder



// clean up our mess

- (void) dealloc
{
    free (points);

    [fabricColor release];
    [backgroundColor release];
    [stitchColor release];
    [textColor release];

    [textFont release];

    [super dealloc];

} // dealloc




// this is how many locations we have.  Each stitch is composed
// of two points - a start and an end (duh...).  The user
// is expected to pick the two points, so it's necessary to
// be able to have an odd number of points in the collection

- (int) count
{
    return (pointCount);
} // count



// add a new stich location to the collection

- (void) addStitchLocation: (BWStitchLocation) stitch
{
    pointCount++;

    // this is very low-traffic function (at most one per click), so a
    // realloc every time is fine.
    points = realloc (points, sizeof(BWStitchLocation) * pointCount);

    points[pointCount - 1] = stitch;

} // addStitchLocation



// get the location at the index

- (BWStitchLocation)  stitchLocationAtIndex: (int) index
{
    // be a little paranoid
    assert (index >= 0 && index < pointCount);

    BWStitchLocation result;
    result = points[index];

    return (result);

} // stitchLocationAtIndex



// lop off the last location.  The UI is planned to only support
// undo for removing of points (no removing stuff from the
// middle of the collection)

- (void) removeLastStitchLocation
{
    pointCount--;

    // be a little paranoid
    assert (pointCount >= 0);

} // removeLastStitchLocation



// someone has changed a visual attribute.  Make sure the view
// (or any other interested parties) knows about it to redraw things.
// saves the controller from having a ton of extra setNeedsDisplay:YES
// scattered around.

- (void) postVisualNotification
{
    NSNotificationCenter *center;
    center = [NSNotificationCenter defaultCenter];

    [center 
	postNotificationName: BWStitchGroup_VisualAttributeChangeNotification
	object: self];
    
} // postVisualNotification



// change the fabric width - how wide to draw the colored grid

- (void) setFabricWidth: (float) width
{
    threadSize = width;
    [self postVisualNotification];
} // setFabricWidth



// getter for the attribute that's been set above

- (float) fabricWidth
{
    return (threadSize);
} // fabricWidth



// change the spacing between the colored grid lines

- (void) setFabricSpacing: (float) newSpacing
{
    spacing = newSpacing;
    [self postVisualNotification];

} // setFabricSpacing



// getter for the attribute that's been set above

- (float) fabricSpacing
{
    return (spacing);
} // fabricSpacing



// change the width of the stitches

- (void) setStitchSize: (float) newStitchSize
{
    stitchSize = newStitchSize;
    [self postVisualNotification];

} // setStitchSize



// getter for the attribute that's been set above

- (float) stitchSize
{
    return (stitchSize);
} // stitchSize



// change the color the grid is drawn with

- (void) setFabricColor: (NSColor *) color
{
    if (fabricColor != color) {
	[fabricColor release];
	fabricColor = [color copy];
	[self postVisualNotification];
    }

} // setFabricColor



// getter for the attribute that's been set above

- (NSColor *) fabricColor
{
    return ([[fabricColor retain] autorelease]);

} // fabricColor



// change the color the stitches are drawn in

- (void) setStitchColor: (NSColor *) color
{
    if (stitchColor != color) {
	[stitchColor release];
	stitchColor = [color copy];
	[self postVisualNotification];
    }

} // setStitchColor



// getter for the attribute that's been set above

- (NSColor *) stitchColor
{
    return ([[stitchColor retain] autorelease]);

} // stitchColor



// change the color used to draw stuff that hasn't been drawn
// in any other color

- (void) setBackgroundColor: (NSColor *) color
{
    if (backgroundColor != color) {
	[backgroundColor release];
	backgroundColor = [color copy];
	[self postVisualNotification];
    }

} // setBackgroundColor


// getter for the attribute that's been set above

- (NSColor *) backgroundColor
{
    return ([[backgroundColor retain] autorelease]);
} // backgroundColor



// change the color the text labels will be drawn with

- (void) setTextColor: (NSColor *) color
{
    if (textColor != color) {
	[textColor release];
	textColor = [color copy];
	[self postVisualNotification];
    }

} // setTextColor



// getter for the attribute that's been set above

- (NSColor *) textColor
{
    return ([[textColor retain] autorelease]);

} // textColor


// change the way the text labels are displayed

- (void) setTextFont: (NSFont *) font
{
    if (textFont != font) {
	[textFont release];
	textFont = [font copy];
	[self postVisualNotification];
    }

} // setTextFont



// getter for the attribute that's been set above

- (NSFont *) textFont
{
    return ([[textFont retain] autorelease]);

} // textFont



// turn on/off the display of even and odd numbers. There
// are four choices: evens only, odds only, evens and odds,
// or display nothing

- (void) setNumberDisplay: (BWStitchGroup_NumberDisplay) display
{
    numberDisplay = display;

} // setNumberDisplay



// getter for the attribute that's been set above

- (BWStitchGroup_NumberDisplay) numberDisplay
{
    return (numberDisplay);
} // numberDisplay




// move the contents of the world around.

- (void) shiftStitchesBy: (int) number  
	       direction: (BWStitchGroup_ShiftDirection) direction
{
    // first figure out how far we need to move each of the
    // stitches in the x and y direction

    int xdelta = 0;
    int ydelta = 0;

    switch (direction) {
    case BWStitchGroup_ShiftUp:
	ydelta = -number;
	break;
    case BWStitchGroup_ShiftDown:
	ydelta = number;
	break;
    case BWStitchGroup_ShiftLeft:
	xdelta = -number;
	break;
    case BWStitchGroup_ShiftRight:
	xdelta = number;
	break;
    default:
	NSLog (@"bad shift direction");
    }

    // walk through all the points and mutate their x and y values

    BWStitchLocation *scan, *stop;
    scan = points;
    stop = scan + pointCount;

    while (scan < stop) {
	scan->x += xdelta;
	scan->y += ydelta;
	scan++;
    }

} // shiftStitchesBy



// set the area the user wants to see on the eventual printout

- (void) setPrintableArea: (BWIntRect) newArea
{
    printableArea = newArea;

} // setPrintableArea



// getter for the attribute that's been set above

- (BWIntRect) printableArea
{
    return (printableArea);

} // printableArea



// zero out the printable area, turing it into an empty rectangle

- (void) clearPrintableArea
{
    printableArea = BWMakeIntRect (0, 0, 0, 0);

} // clearPrintableArea



// determine min/max of the X's and Y's in the collection.
// pretty much calculate a default printable area.

- (BWIntRect) contentBounds
{
    BWIntRect result;

    if (pointCount == 0) {
	// if no points, the content bounds is empty
	result = BWMakeIntRect (0, 0, 0, 0);

    } else {
	// otherwise, walk the collection and do the CS-101
	// min/max thing

	int minX = INT_MAX, maxX = 0;
	int minY = INT_MAX, maxY = 0;
	
	BWStitchLocation *scan, *stop;
	scan = points;
	stop = points + pointCount;
	
	while (scan < stop) {
	    minX = MIN (minX, scan->x);
	    maxX = MAX (maxX, scan->x);
	    
	    minY = MIN (minY, scan->y);
	    maxY = MAX (maxY, scan->y);
	    
	    scan++;
	}
	
	result = BWMakeIntRect (minX, minY,
				maxX - minX, maxY - minY);
    }	

    return (result);

} // contentBounds



@end // BWStitchGroup


