// BWStitchNoteDocument -- this is the controller class for the StitchNotator

#import "BWStitchNoteDocument.h"
#import "BWDocument-Undo.h"
#import "BWStitchGroup.h"
#import "BWFabricView.h"

@interface BWStitchNoteDocument(Private)
- (void) setupContextHelp;
@end // Private

@implementation BWStitchNoteDocument 


// do one-time initializations

- (id) init
{
    if (self = [super init]) {
	group = [[BWStitchGroup alloc] init];

	[self setupUndo];
    }

    return (self);

} // init



// clean up our mess

- (void) dealloc
{
    // clean up any undo-related storage
    [self deallocUndo];

    // dispose of our model
    [group release];

    [super dealloc];

} // dealloc



// synchronize the UI items in the drawer with the values that
// are in the model
- (void) updateUI
{
    // update the sliders
    [fabricWidthSlider setFloatValue: [group fabricWidth]];
    [fabricSpacingSlider setFloatValue: [group fabricSpacing]];
    [stitchSizeSlider setFloatValue: [group stitchSize]];

    // update the color wells
    [fabricColorWell setColor: [group fabricColor]];
    [backgroundColorWell setColor: [group backgroundColor]];
    [stitchColorWell setColor: [group stitchColor]];
    [textColorWell setColor: [group textColor]];

    // update the font and font size menus
    NSFont *font;
    font = [group textFont];

    NSString *familyName;
    familyName = [font familyName];

    int pointSize;
    pointSize = (int)[font pointSize];

    int index;
    index = [sizeMenuButton indexOfItemWithTag: pointSize];
    [sizeMenuButton selectItemAtIndex: index];
    
    index = [fontMenuButton indexOfItemWithTitle: familyName];
    [fontMenuButton selectItemAtIndex: index];

    NSFontManager *fontManager;
    fontManager = [NSFontManager sharedFontManager];

    NSFontTraitMask traits;
    traits = [fontManager traitsOfFont: font];

    [boldCheckbox setState: NSOffState];
    [italicCheckbox setState: NSOffState];

    if (traits & NSBoldFontMask) {
	[boldCheckbox setState: NSOnState];
    }
    if (traits & NSItalicFontMask) {
	[italicCheckbox setState: NSOnState];
    }

#if 0
    // !!! This doesn't work reliably.  Like, "Monaco", if you try to
    // !!! create an italic version, it doesn't really (the created
    // !!! font doesn't have the italic trait.  The font does say it
    // !!! has traits for italic, but it lies.

    // if the font doesn't support the given trait, disable the checkbox
    // the traits are kinda backwards from what I expected.  Rather
    // than asking "this font family supports bold", you ask
    // "does this font have the "sorry, I can't do italics" property

    if ([fontManager fontNamed: familyName  hasTraits: NSUnboldFontMask]) {
	[boldCheckbox setEnabled: YES];
    } else {
	[boldCheckbox setEnabled: NO];
    }

    if ([fontManager fontNamed: familyName  hasTraits: NSUnitalicFontMask]) {
	[italicCheckbox setEnabled: YES];
    } else {
	[italicCheckbox setEnabled: NO];
    }
#else
    // !!! so instead, usea  sledgehammer and make a font with both
    // !!! traits and then check to resulting font
    // !!! I don't like this since it can cause delays when browsing
    // !!! through fonts

    font = [fontManager fontWithFamily: familyName
			traits: NSItalicFontMask | NSBoldFontMask
			weight: 5
			size: pointSize];
    traits = [fontManager traitsOfFont: font];
    
    if (traits & NSBoldFontMask) {
	[boldCheckbox setEnabled: YES];
    } else {
	[boldCheckbox setEnabled: NO];
    }

    if (traits & NSItalicFontMask) {
	[italicCheckbox setEnabled: YES];
    } else {
	[italicCheckbox setEnabled: NO];
    }
#endif

    index = [numberDisplayButton
		indexOfItemWithTag: [group numberDisplay]];
    [numberDisplayButton selectItemAtIndex: index];

} // updateUI



// the name of the nib file that describes the document window

- (NSString *) windowNibName
{
    return (@"BWStitchNoteDocument");
} // windowNibName


// create the contents of the font-name button

- (void) populateFontButton
{
    // the names of the font family, sorted.
    NSArray *fontNames;
    fontNames = [[NSFontManager sharedFontManager] availableFontFamilies];
    fontNames = [fontNames sortedArrayUsingSelector: @selector(compare:)];

    // get the menu that belongs to the font button
    NSMenu *fontMenu;
    fontMenu = [fontMenuButton menu];

    // out with the old
    while ([fontMenu numberOfItems] > 0) {
	[fontMenu removeItemAtIndex: 0];
    }

    // in with the new
    NSEnumerator *enumerator;
    NSString *name;

    enumerator = [fontNames objectEnumerator];
    while (name = [enumerator nextObject]) {

	// there are some 'noise' fonts that end with TT or LET which apps
	// like mail.app don't show.  So filter those out, as well as those
	// that have a leading dot in the name
	if ([name characterAtIndex: 0] == '.') {
	    continue;
	}

	NSRange range;
	range = [name rangeOfString: @"TT" 
		      options: NSBackwardsSearch | NSAnchoredSearch];
	if (range.location != NSNotFound) {
	    continue;
	}

	range = [name rangeOfString: @"LET" 
		      options: NSBackwardsSearch | NSAnchoredSearch];
	if (range.location != NSNotFound) {
	    continue;
	}

	// if we survived down to here, we have a worthwhile font
	[fontMenu addItemWithTitle: name  action: nil  keyEquivalent: @""];
    }

} // populateFontButton



// the nib for our window has loaded.  Do any manipulations on connections
// made in the nib

- (void) windowControllerDidLoadNib: (NSWindowController *) windowController
{
    // tell the view what to use for drawing information
    [fabricView setStitchGroup: group];

    // add observers on the view to react to some user actions
    NSNotificationCenter *center;
    center = [NSNotificationCenter defaultCenter];
    
    [center addObserver: self
	    selector: @selector(newStitchNotification:)
	    name: BWFabricViewNewStitchNotification
	    object: fabricView];

    [center addObserver: self
	    selector: @selector(printableAreaNotification:)
	    name: BWFabricViewPrintableAreaNotification
	    object: fabricView];

    [self populateFontButton];

    [self setupContextHelp];

} // windowControllerDidLoadNib



// create an NSData out of the document so it can be saved to a file

- (NSData *) dataRepresentationOfType: (NSString *) aType
{
    NSData *data;
    data = [NSArchiver archivedDataWithRootObject: group];

    return (data);

} // dataRepresentationOfType



// starting with an NSData of the document (created above in
// dataRepresentationOfType), create the document.  That is,
// pull out the model and replace our current model with the
// one from the NSData

- (BOOL) loadDataRepresentation: (NSData *) data 
			 ofType: (NSString *) aType
{
    // disconnect the view
    [fabricView setStitchGroup: nil];

    // replace the moedl
    [group release];
    group = [NSUnarchiver unarchiveObjectWithData: data];
    [group retain];

    // reconnect the view to the new model
    [fabricView setStitchGroup: group];

    // stash away current values of some attributes
    [self rememberUndoInformation];

    return (YES);

} // loadDataRepresentation



// notification from the view.  The user has clicked on the grid
// to place a new stitch location

- (void) newStitchNotification: (NSNotification *) notification
{
    NSValue *location;
    location = (NSValue*)[notification userInfo]; // hack!
    // the location is stored directly in the userInfo value, rather
    // than traditionally in a dictionary

    BWStitchLocation loc;
    
    [location getValue: &loc];

    [self addStitchLocation: loc];

} // newStitchNotification



// notification from the view.  The user has specified the printable
// area

- (void) printableAreaNotification: (NSNotification *) notification
{
    BWIntRect rect;

    rect = [fabricView getTrackingPrintArea];

    [self setPrintableArea: rect];
    [self updateUI];

} // printableAreaNotification



// open and close the drawer

- (IBAction) toggleDrawer: (id) sender
{
    [prefsDrawer toggle: sender];

    // update the drawer's controls to match the reality of the world
    [self updateUI]; 

} // toggleDrawer



// slider action for changing the fabric width, which is the
// size of the grid being drawn

- (IBAction) changeFabricWidth: (id) sender
{
    // change the value
    [group setFabricWidth: [sender floatValue]];
    
    // enqueue an undo notification for when the dragging's done
    [self enqueueFabricWidthChange: sender];

} // changeFabricWidth



// slider action for changing the spacing between fabric lines

- (IBAction) changeFabricSpacing: (id) sender
{
    // change the value
    [group setFabricSpacing: [sender floatValue]];

    // enqueue an undo notification for when the dragging's done
    [self enqueueFabricSpacingChange: sender];

} // changeFabricSpacing



// slider action for changing the size of the stitch ovals

- (IBAction) changeStitchSize: (id) sender
{
    // change the value
    [group setStitchSize: [sender floatValue]];

    // enqueue an undo notification for when the dragging's done
    [self enqueueStitchSizeChange: sender];

} // changeStitchSize



// color well action for changing the grid color

- (IBAction) changeFabricColor: (id) sender
{
    // change the value
    [group setFabricColor: [sender color]];

    // enqueue an undo notification for when the dragging's done
    [self enqueueFabricColorChange: sender];

} // changeFabricColor



// color well action for changing the background color
// (what's drawn where nothing else is drawn)

- (IBAction) changeBackgroundColor: (id) sender
{
    // change the value
    [group setBackgroundColor: [sender color]];

    // enqueue an undo notification for when the dragging's done
    [self enqueueBackgroundColorChange: sender];

} // changeBackgroundColor



// color well action for changing the color the stitch ovals
// are drawn with

- (IBAction) changeStitchColor: (id) sender
{
    // change the value
    [group setStitchColor: [sender color]];

    // enqueue an undo notification for when the dragging's done
    [self enqueueStitchColorChange: sender];

} // changeStitchColor



// color well action for changing the color of the text labels

- (IBAction) changeTextColor: (id) sender
{
    // change the value
    [group setTextColor: [sender color]];

    // enqueue an undo notification for when the dragging's done
    [self enqueueTextColorChange: sender];

} // changeTextColor



// popup menu action for font or size changes

- (IBAction) changeFont: (id) sender
{
    NSString *familyName;
    familyName = [fontMenuButton titleOfSelectedItem];

    int fontSize;
    fontSize = [[sizeMenuButton selectedItem] tag];

    BOOL bold;
    bold = [boldCheckbox state] == NSOnState;

    BOOL italic;
    italic = [italicCheckbox state] == NSOnState;

    // registers undo information, and then changes the model
    [self setFontFamily: familyName  size: fontSize
	  bold: bold  italic: italic];

} // changeFont



// popup menu action for changing which labels get shown

- (IBAction) changeNumberDisplay: (id) sender;
{
    // the tag value matches the enum values specified in
    // BWStitchGroup.h
    BWStitchGroup_NumberDisplay value;
    value = [[sender selectedItem] tag];

    // registers undo information, then changes the model
    [self setNumberDisplay: value];

} // changeOddOnly



// action for moving stiches around in the canvas

- (IBAction) shiftStitches: (id) sender
{
    // look at the tag for shift.  The tag value matches the
    // the enum values specified in BWStitchGroup.h
    int tag = [sender tag];

    // registers undo information, then changes the model
    [self shiftStitchesBy: 1
	  direction: (BWStitchGroup_ShiftDirection) tag];

} // shiftStitches



// action for starting the tracking of printable area

- (IBAction) selectPrintableArea: (id) sender
{
    // let the view handle the dirty work of figuring out what the
    // user wants

    [fabricView startTrackingPrintableArea];

} // selectPrintableArea



// action for zeroing out the printable area (after which we'll try
// to figure out the best size based on what the user has drawn

- (IBAction) clearPrintableArea: (id) sender
{
    [self clearPrintableArea]; // this is over in the undo code
    [fabricView setNeedsDisplay: YES];

} // clearPrintableArea



// what is our window?  Dig into the window controllers to try to
// figure this out.

- (NSWindow *) window
{
    NSArray *windowControllers;
    windowControllers = [self windowControllers];

    NSWindowController *controller;
    controller = [windowControllers objectAtIndex: 0];

    NSWindow *window;
    window = [controller window];

    return (window);

} // window



// print the document

- (void) printDocument: (id) sender
{
    // make a special view to do the printing
    BWFabricView *printView;
    printView = [[BWFabricView alloc] initForPrintingWithGroup: group];

    NSPrintOperation *printOperation;
    printOperation = [NSPrintOperation 
			 printOperationWithView: printView
			 printInfo: [self printInfo]];

    [printOperation runOperationModalForWindow: [self window]
		    delegate: self
		    didRunSelector:
			@selector(printOperationDidRun:success:contextInfo:)
		    contextInfo: NULL];

    // the view gets retained by the printOperation
    [printView release];

} // print



// this is the didRunSelector for a print operation

- (void) printOperationDidRun: (NSPrintOperation *) printOperation
		      success: (BOOL) success
		  contextInfo: (void *) info
{
    if (success) {
	// store printInfo(e) if you want
	// printInfo = [printOperation printInfo];
	// although docs say it best not to save it, so that
	// the appropriate defaults are used each time
    }

} // printOperationDidRun



// the user has chosen Copy.  Put a pdf on the clipboard of what we
// would have printed

- (IBAction) copy: (id) sender
{
    // make a place for the PDF data to be stored
    NSMutableData *data;
    data = [NSMutableData dataWithLength: 0];

    // making a printing view
    BWFabricView *printView;
    printView = [[BWFabricView alloc] initForPrintingWithGroup: group];

    // run the print operation, accumulating the data into the
    // NSData created above
    NSPrintOperation *printOperation;
    printOperation = [NSPrintOperation 
			 PDFOperationWithView: printView
			 insideRect: [printView bounds]
			 toData: data
			 printInfo: [self printInfo]];

    [printOperation runOperation];

    // now we have the PDF data.  stick it on the pasteboard
    NSPasteboard *pasteboard;
    pasteboard = [NSPasteboard  generalPasteboard];

    [pasteboard declareTypes: [NSArray arrayWithObject: NSPDFPboardType]
	       owner: self];

    [pasteboard setData: data
		forType: NSPDFPboardType];

    [printView release];

} // copy



// enable / disable the Clear Printable Area menu item, based on
// whether we have a user-chosen printable area

- (BOOL) validateMenuItem: (NSMenuItem *) anItem
{
    BOOL result = YES;

    if ([anItem action] == @selector(clearPrintableArea:)) {
	BWIntRect printableArea;
	printableArea = [group printableArea];
	
	if (printableArea.size.width <= 0
	    || printableArea.size.height <= 0) {
	    result = NO;
	}
	
    } else {
	result = [super validateMenuItem: anItem];
    }

    return (result);

} // validateMenuItem


// --------------------------------------------------
// Help stuff

// set up context help, which is between tooltips and the heavy-weight
// HelpView help.

- (void) setupContextHelp
{
    NSHelpManager *helpManager;
    helpManager = [NSHelpManager sharedHelpManager];

    NSBundle *mainBundle;
    mainBundle = [NSBundle mainBundle];

    [helpManager
	setContextHelp: [mainBundle
			    contextHelpForKey: @"fabricviewContextHelp.rtfd"]
	forObject: fabricView];

    [helpManager
	setContextHelp: [mainBundle
			    contextHelpForKey: @"drawerdisclosureContextHelp.rtfd"]
	forObject: drawerDisclosureButton];

    [helpManager
	setContextHelp: [mainBundle
			    contextHelpForKey: @"fabricwidthContextHelp.rtfd"]
	forObject: fabricWidthSlider];

    [helpManager
	setContextHelp: [mainBundle
			    contextHelpForKey: @"fabricspacingContextHelp.rtfd"]
	forObject: fabricSpacingSlider];

    [helpManager
	setContextHelp: [mainBundle
			    contextHelpForKey: @"stitchsizeContextHelp.rtfd"]
	forObject: stitchSizeSlider];

    [helpManager
	setContextHelp: [mainBundle
			    contextHelpForKey: @"fabriccolorContextHelp.rtfd"]
	forObject: fabricColorWell];

    [helpManager
	setContextHelp: [mainBundle
			    contextHelpForKey: @"backgroundcolorContextHelp.rtfd"]
	forObject: backgroundColorWell];

    [helpManager
	setContextHelp: [mainBundle
			    contextHelpForKey: @"stitchcolorContextHelp.rtfd"]
	forObject: stitchColorWell];

    [helpManager
	setContextHelp: [mainBundle
			    contextHelpForKey: @"textcolorContextHelp.rtfd"]
	forObject: textColorWell];

    [helpManager
	setContextHelp: [mainBundle
			    contextHelpForKey: @"fontContextHelp.rtfd"]
	forObject: fontMenuButton];

    [helpManager
	setContextHelp: [mainBundle
			    contextHelpForKey: @"fontContextHelp.rtfd"]
	forObject: sizeMenuButton];

    [helpManager
	setContextHelp: [mainBundle
			    contextHelpForKey: @"fontContextHelp.rtfd"]
	forObject: boldCheckbox];

    [helpManager
	setContextHelp: [mainBundle
			    contextHelpForKey: @"fontContextHelp.rtfd"]
	forObject: italicCheckbox];

    [helpManager
	setContextHelp: [mainBundle
			    contextHelpForKey: @"numberdisplayContextHelp.rtfd"]
	forObject: numberDisplayButton];

} // setupContextHelp


@end // BWStitchNoteDocument

